#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#include <bm.h>

#include <internal.h>

#include <cryptsetup-java.h>

#define min(a,b) ((a)<(b)?(a):(b))

static int and_use_fds = 0;
static struct and_device* and_fds[AND_MAX_FD] = {NULL};

enum lock_type {
	DEV_LOCK_NONE = 0,
	DEV_LOCK_READ,
	DEV_LOCK_WRITE
};

struct device {
};

size_t size_round_up(size_t size, size_t block)
{
	size_t s = (size + (block - 1)) / block;
	return s * block;
}

int device_alloc(struct crypt_device *cd, struct device **device, const char *path)
{
	*device = (struct device*)path;
	return 0;
}

void device_free(struct crypt_device *cd, struct device *d)
{
	struct and_device *device = (struct and_device*)d;
	if (!device)
		return;

	if (device->fd >= 0 && device->fd < AND_MAX_FD)
	{
		log_dbg(cd, "Device FD %d close", device->fd);
		
		and_fds[device->fd] = NULL;
		and_use_fds &= ~(1 << device->fd);
		device->fd = -1;
	}
}

/* Get data size in bytes */
int device_size(struct device *d, uint64_t *size)
{
	struct and_device *device = (struct and_device*)d;
	if (device->blk->ioctl(device->blk, IOCTL_GET_SIZE, size))
		return -EINVAL;
	else
		return 0;
}

size_t device_block_size(struct crypt_device *cd, struct device *d)
{
	struct and_device *device = (struct and_device*)d;
	if (!device)
		return 0;

	if (device->block_size)
		return device->block_size;

	uint32_t size; 
	if (!device->blk->ioctl(device->blk, IOCTL_GET_SECTOR_SIZE, &size))
		device->block_size = size;

	if (!device->block_size)
		log_dbg(cd, "Cannot get block size for device");

	return device->block_size;
}

size_t device_optimal_encryption_sector_size(struct crypt_device *cd, struct device *d)
{
	uint32_t size; 
	struct and_device *device = (struct and_device*)d;
	if (device && !device->blk->ioctl(device->blk, IOCTL_GET_SECTOR_SIZE, &size))
		return size;
	else
		return SECTOR_SIZE;
}

/* Get path to device / file */
const char *device_path(const struct device *device)
{
	return "/dev";
}

/* Get block device path */
const char *device_block_path(const struct device *device)
{
	return NULL;
}

int device_block_adjust(struct crypt_device *cd,
			struct device *d,
			enum devcheck device_check,
			uint64_t device_offset,
			uint64_t *size,
			uint32_t *flags)
{
	uint8_t real_readonly;
	uint64_t real_size;

	struct and_device *device = (struct and_device*)d;
	if (!device)
		return -ENOTBLK;

	if (device->blk->ioctl(device->blk, IOCTL_GET_SIZE, &real_size))
		return -EINVAL;
	real_size >>= SECTOR_SHIFT;
	if (device->blk->ioctl(device->blk, IOCTL_GET_IS_RO, &real_readonly))
		return -EINVAL;

	if (device_offset >= real_size) {
		log_err(cd, _("Requested offset is beyond real size of device"));
		return -EINVAL;
	}

	if (size && !*size)
	{
		*size = real_size;
		if (!*size)
		{
			log_err(cd, _("Device has zero size"));
			return -ENOTBLK;
		}
		*size -= device_offset;
	}

	/* in case of size is set by parameter */
	if (size && ((real_size - device_offset) < *size))
	{
		log_dbg(cd, "Device: offset = %" PRIu64 " requested size = %" PRIu64
			", backing device size = %" PRIu64,
			device_offset, *size, real_size);
		log_err(cd, _("Device is too small."));
		return -EINVAL;
	}

	if (flags && real_readonly)
		*flags |= CRYPT_ACTIVATE_READONLY;

	if (size)
		log_dbg(cd, "Calculated device size is %" PRIu64" sectors (%s), offset %" PRIu64 ".",
			*size, real_readonly ? "RO" : "RW", device_offset);
	return 0;
}

/* Get device-mapper name of device (if possible) */
const char *device_dm_name(const struct device *device)
{
	return NULL;
}

int device_is_identical(struct device *d1, struct device *d2)
{
	if (!d1 || !d2)
		return 0;

	struct and_device *device1 = (struct and_device*)d1;
	struct and_device *device2 = (struct and_device*)d2;
	if (device1 == device2 || device1->blk == device2->blk)
		return 1;

	return 0;
}

int device_write_lock(struct crypt_device *cd, struct device *d)
{
	if (!crypt_metadata_locking_enabled())
		return 0;

	struct and_device *device = (struct and_device*)d;
	assert(!device_is_locked(device));

	device->lock = DEV_LOCK_WRITE;
	return 0;
}

int device_read_lock(struct crypt_device *cd, struct device *d)
{
	if (!crypt_metadata_locking_enabled())
		return 0;

	struct and_device *device = (struct and_device*)d;
	device->lock = DEV_LOCK_READ;
	return 0;
}

void device_read_unlock(struct crypt_device *cd, struct device *d)
{
	if (!crypt_metadata_locking_enabled())
		return;

	struct and_device *device = (struct and_device*)d;
	assert(device_is_locked(device));

	device->lock = DEV_LOCK_NONE;
}

void device_write_unlock(struct crypt_device *cd, struct device *d)
{
	if (!crypt_metadata_locking_enabled())
		return;

	struct and_device *device = (struct and_device*)d;
	assert(device_is_locked(device) && !(device && device->lock == DEV_LOCK_READ));

	device->lock = DEV_LOCK_NONE;
}

bool device_is_locked(struct device *d)
{
	struct and_device *device = (struct and_device*)d;
	return (device && device->lock);
}

size_t device_alignment(struct device *d)
{
	return DEFAULT_MEM_ALIGNMENT;
}

int device_direct_io(const struct device *d)
{
	return 1;
}

void device_disable_direct_io(struct device *d)
{
	/* NOP */
}

void device_set_block_size(struct device *d, size_t size)
{
	struct and_device *device = (struct and_device*)d;
	if (!device)
		return;

	uint32_t sector; 
	if (device->blk->ioctl(device->blk, IOCTL_GET_SECTOR_SIZE, &sector))
		return;

	if ((size%sector) == 0)
		device->block_size = size;
}

/*
 * Common wrapper for device sync.
 */
void device_sync(struct crypt_device *cd, struct device *d)
{
	// No need
}

/* For a file, allocate the required space */
int device_fallocate(struct device *d, uint64_t size)
{
	return -EINVAL;
}

int device_check_size(struct crypt_device *cd,
		      struct device *d,
		      uint64_t req_offset, int falloc)
{
	uint64_t dev_size;

	if (device_size(d, &dev_size))
	{
		log_dbg(cd, "Cannot get device size for device");
		return -EIO;
	}

	log_dbg(cd, "Device size %" PRIu64 ", offset %" PRIu64 ".", dev_size, req_offset);

	if (req_offset > dev_size)
	{
		/* If it is header file, increase its size */
		if (falloc && !device_fallocate(d, req_offset))
			return 0;

		log_err(cd, _("Device is too small. Need at least %" PRIu64 " bytes."),
			req_offset);
		return -EINVAL;
	}

	return 0;
}

static int device_open_internal(struct crypt_device *cd, struct and_device *device)
{
	device->offset = 0;
	if (device->fd >= 0)
		return device->fd;

	int fd;
	for (fd=0; fd<AND_MAX_FD; fd++)
		if ((and_use_fds&(1<<fd)) == 0)
			break;
	if (fd >= AND_MAX_FD)
	{
		log_err(cd, _("Free FD not found"));
		return -1;
	}

	and_use_fds |= (1<<fd);
	device->fd = fd;
	and_fds[fd] = device;
	return fd;
}

int device_open(struct crypt_device *cd, struct device *d, int flags)
{
	struct and_device *device = (struct and_device*)d;
	assert(!device_is_locked(device));
	return device_open_internal(cd, device);
}

int device_open_locked(struct crypt_device *cd, struct device *d, int flags)
{
	struct and_device *device = (struct and_device*)d;
	assert(!crypt_metadata_locking_enabled() || device_is_locked(device));

	if (device->lock == DEV_LOCK_WRITE)
		return -1;

	int access = flags & O_ACCMODE;
	if (access == O_WRONLY)
		access = O_RDWR;

	if (access != O_RDONLY && device->lock == DEV_LOCK_READ)
		return -1;

	return device_open_internal(cd, device);
}

ssize_t read_buffer(int fd, void *buf, size_t length)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;
	ssize_t read_size = blk->pread(blk, buf, length, and_fds[fd]->offset);
	if (read_size > 0)
		and_fds[fd]->offset += read_size;
	return read_size;
}

ssize_t read_lseek_blockwise(int fd, size_t bsize, size_t alignment,
			     void *buf, size_t length, off_t offset)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;
	if (!buf || bsize <= 0)
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;

	if (offset < 0)
	{
		uint64_t size;
		if (blk->ioctl(blk, IOCTL_GET_SECTOR_SIZE, &size))
			return -1;
		offset = and_fds[fd]->offset = size - offset;
	} else
		and_fds[fd]->offset = offset;

	uint32_t sector_size = 512;
	blk->ioctl(blk, IOCTL_GET_SECTOR_SIZE, &sector_size);

	uint64_t sector = (and_fds[fd]->offset/sector_size)*sector_size;
	size_t read, bytes = and_fds[fd]->offset%sector_size;
	if (bytes > 0)
	{
		uint8_t tmp[sector_size];
		if (blk->pread(blk, tmp, sector_size, sector) <= 0)
			return -1;
		sector+=sector_size;
		
		read = min(sector_size-bytes, length);
		memcpy(buf, tmp+bytes, read);
	} else
		read = 0;

	while (read < length)
	{
		size_t need = length-read;
		if (need >= sector_size)
		{
			if (blk->pread(blk, buf+read, sector_size, sector) <= 0)
				break;
			sector+=sector_size;
			read += sector_size;
		} else
		{
			uint8_t tmp[sector_size];
			if (blk->pread(blk, tmp, sector_size, sector) <= 0)
				break;
			memcpy(buf+read, tmp, need);
			read += need;
		}
	}

	if (read > 0)
		and_fds[fd]->offset += read;
	return read;
}

ssize_t read_blockwise(int fd, size_t bsize, size_t alignment,
		       void *buf, size_t length)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;
	ssize_t read_size = blk->pread(blk, buf, length, and_fds[fd]->offset);
	if (read_size > 0)
		and_fds[fd]->offset += read_size;
	return read_size;
}

ssize_t write_buffer(int fd, const void *buf, size_t length)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;
	ssize_t write_size = blk->pwrite(blk, buf, length, and_fds[fd]->offset);
	if (write_size > 0)
		and_fds[fd]->offset += write_size;
	return write_size;
}

ssize_t write_lseek_blockwise(int fd, size_t bsize, size_t alignment,
			     void *buf, size_t length, off_t offset)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;
	if (!buf || bsize <= 0)
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;

	if (offset < 0)
	{
		uint64_t size;
		if (blk->ioctl(blk, IOCTL_GET_SECTOR_SIZE, &size))
			return -1;
		offset = and_fds[fd]->offset = size - offset;
	} else
		and_fds[fd]->offset = offset;

	ssize_t write_size = blk->pwrite(blk, buf, length, and_fds[fd]->offset);
	if (write_size > 0)
		and_fds[fd]->offset += write_size;
	return write_size;
}

ssize_t write_blockwise(int fd, size_t bsize, size_t alignment,
			void *buf, size_t length)
{
	if (fd < 0 || fd >= AND_MAX_FD || !and_fds[fd])
		return -EINVAL;

	block_t *blk = and_fds[fd]->blk;
	ssize_t write_size = blk->pwrite(blk, buf, length, and_fds[fd]->offset);
	if (write_size > 0)
		and_fds[fd]->offset += write_size;
	return write_size;
}

// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void device_close(struct crypt_device *cd, struct device *device)
{
	/* NOP */
}

int device_open_excl(struct crypt_device *cd, struct device *device, int flags)
{
	return -ENOTSUP;
}
