#include "crypto_mbedtls.h"

#include <string.h>
#include <errno.h>
#include <stdlib.h>

#include <mbedtls/aes.h>
#include <mbedtls/sha1.h>
#include <mbedtls/sha256.h>
#include <mbedtls/sha512.h>
#include <mbedtls/ripemd160.h>
#include <mbedtls/serpent.h>
#include <mbedtls/twofish.h>
#include <mbedtls/blowfish.h>
#include <mbedtls/camellia.h>

#include <mbedtls/pkcs5.h>
#include <mbedtls/md.h>
#include <mbedtls/ccm.h>

#include <mbedtls/version.h>

#include "crypto_backend_internal.h"

#ifdef USE_BM_BACKEND
#include <bm_cipher.h>
#else
struct crypt_cipher {
	void *hd_enc;
	void *hd_dec;
	int chiper_type;
	size_t iv_length;
};
#endif

struct crypt_hash {
	void *md;
	int hash_id;
	int hash_len;
};

int crypt_backend_init(bool fips)
{
	return 0;
}

void crypt_backend_destroy(void)
{
}

uint32_t crypt_backend_flags(void)
{
	return 0;
}

const char *crypt_backend_version(void)
{
	return MBEDTLS_VERSION_STRING_FULL;
}

/* HASH */
int crypt_hash_type(const char *hash_spec)
{
	if (strcmp(hash_spec, "sha1") == 0)
		return MBEDTLS_MD_SHA1;
	else if (strcmp(hash_spec, "sha256") == 0)
		return MBEDTLS_MD_SHA256;
	else if (strcmp(hash_spec, "sha512") == 0)
		return MBEDTLS_MD_SHA512;
	else if (strcmp(hash_spec, "ripemd160") == 0)
		return MBEDTLS_MD_RIPEMD160;
	else
		return MBEDTLS_MD_NONE;
}

static inline int _crypt_hash_size(int hash_id)
{
	switch (hash_id)
	{
	case MBEDTLS_MD_SHA1: return HASH_SHA1_SIZE;
	case MBEDTLS_MD_SHA256: return HASH_SHA256_SIZE;
	case MBEDTLS_MD_SHA512: return HASH_SHA512_SIZE;
	case MBEDTLS_MD_RIPEMD160: return HASH_RIPE160_SIZE;
	}
	return -EINVAL;
}

int crypt_hash_size(const char *name)
{
	const int hash_id = crypt_hash_type(name);
	if (!hash_id)
		return -EINVAL;

	return _crypt_hash_size(hash_id);
}

int crypt_hash_init(struct crypt_hash **ctx, const char *name)
{
	struct crypt_hash *h;

	h = malloc(sizeof(*h));
	if (!h)
		return -ENOMEM;

	h->hash_id = crypt_hash_type(name);
	if (!h->hash_id)
	{
		free(h);
		return -EINVAL;
	}

	int ret;
	switch (h->hash_id)
	{
	case MBEDTLS_MD_SHA1:
		h->md = malloc(sizeof(mbedtls_sha1_context));
		mbedtls_sha1_init(h->md);
		ret = mbedtls_sha1_starts_ret(h->md);
		break;
	case MBEDTLS_MD_SHA256:
		h->md = malloc(sizeof(mbedtls_sha256_context));
		mbedtls_sha256_init(h->md);
		ret = mbedtls_sha256_starts_ret(h->md, 0);
		break;
	case MBEDTLS_MD_SHA512:
		h->md = malloc(sizeof(mbedtls_sha512_context));
		mbedtls_sha512_init(h->md);
		ret = mbedtls_sha512_starts_ret(h->md, 0);
		break;
	case MBEDTLS_MD_RIPEMD160:
		h->md = malloc(sizeof(mbedtls_ripemd160_context));
		mbedtls_ripemd160_init(h->md);
		ret = mbedtls_ripemd160_starts_ret(h->md);
		break;
	default: ret =  -EINVAL; break;
	}

	if (ret)
	{
		crypt_hash_destroy(h);
		return ret;
	}

	h->hash_len = _crypt_hash_size(h->hash_id);
	*ctx = h;
	return 0;
}

static int crypt_hash_restart(struct crypt_hash *ctx)
{
	int ret;
	switch (ctx->hash_id)
	{
	case MBEDTLS_MD_SHA1: ret = mbedtls_sha1_starts_ret(ctx->md); break;
	case MBEDTLS_MD_SHA256: ret = mbedtls_sha256_starts_ret(ctx->md, 0); break;
	case MBEDTLS_MD_SHA512: ret = mbedtls_sha512_starts_ret(ctx->md, 0); break;
	case MBEDTLS_MD_RIPEMD160: ret = mbedtls_ripemd160_starts_ret(ctx->md); break;
	default: ret = -EINVAL; break;
	}

	return ret;
}

int crypt_hash_write(struct crypt_hash *ctx, const char *buffer, size_t length)
{
	switch (ctx->hash_id)
	{
	case MBEDTLS_MD_SHA1: mbedtls_sha1_update_ret(ctx->md, (const unsigned char *)buffer, length); break;
	case MBEDTLS_MD_SHA256: mbedtls_sha256_update_ret(ctx->md, (const unsigned char *)buffer, length); break;
	case MBEDTLS_MD_SHA512: mbedtls_sha512_update_ret(ctx->md, (const unsigned char *)buffer, length); break;
	case MBEDTLS_MD_RIPEMD160: mbedtls_ripemd160_update_ret(ctx->md, (const unsigned char *)buffer, length); break;
	default: return -EINVAL;
	}

	return 0;
}

int crypt_hash_final(struct crypt_hash *ctx, char *buffer, size_t length)
{
	unsigned char hash[64];

	if (length > (size_t)ctx->hash_len)
		return -EINVAL;

	int ret;
	switch (ctx->hash_id)
	{
	case MBEDTLS_MD_SHA1: ret = mbedtls_sha1_finish_ret(ctx->md, hash); break;
	case MBEDTLS_MD_SHA256: ret = mbedtls_sha256_finish_ret(ctx->md, hash); break;
	case MBEDTLS_MD_SHA512: ret = mbedtls_sha512_finish_ret(ctx->md, hash); break;
	case MBEDTLS_MD_RIPEMD160: ret = mbedtls_ripemd160_finish_ret(ctx->md, hash); break;
	default: ret = -EINVAL; break;
	}
	if (ret)
		return -EINVAL;

	memcpy(buffer, hash, length);
	crypt_hash_restart(ctx);

	return 0;
}

void crypt_hash_destroy(struct crypt_hash *ctx)
{
	switch (ctx->hash_id)
	{
	case MBEDTLS_MD_SHA1: mbedtls_sha1_free(ctx->md); break;
	case MBEDTLS_MD_SHA256: mbedtls_sha256_free(ctx->md); break;
	case MBEDTLS_MD_SHA512: mbedtls_sha512_free(ctx->md); break;
	case MBEDTLS_MD_RIPEMD160: mbedtls_ripemd160_free(ctx->md); break;
	}
	free(ctx->md);
	free(ctx);
}

/* HMAC */
int crypt_hmac_size(const char *name)
{
	return crypt_hash_size(name);
}

/* RNG */
int crypt_backend_rng(char *buffer, size_t length, int quality, int fips)
{
	return -EINVAL;
}

/* PBKDF */
int crypt_pbkdf(const char *kdf, const char *hash,
		const char *password, size_t password_length,
		const char *salt, size_t salt_length,
		char *key, size_t key_length,
		uint32_t iterations, uint32_t memory, uint32_t parallel)

{
	int hash_id;
	int ret;

	if (!kdf)
		return -EINVAL;

	if (!strcmp(kdf, "pbkdf2")) {
		hash_id = crypt_hash_type(hash);
		if (!hash_id)
			return -EINVAL;

		const mbedtls_md_info_t *info = mbedtls_md_info_from_type(hash_id);
		if (!info)
			return -EINVAL;
		
		mbedtls_md_context_t ctx;
		mbedtls_md_init(&ctx);

		ret = mbedtls_md_setup(&ctx, info, 1);
		if (!ret)
		{
			ret = mbedtls_pkcs5_pbkdf2_hmac(&ctx, (const unsigned char *)password,
                       password_length, (const unsigned char *)salt, salt_length, iterations,
                       key_length, (unsigned char *)key);
		}

		mbedtls_md_free(&ctx);

		if (ret)
			return -EINVAL;
		return 0;
	} else if (!strncmp(kdf, "argon2", 6)) {
		return argon2(kdf, password, password_length, salt, salt_length,
			      key, key_length, iterations, memory, parallel);
	}

	return -EINVAL;
}

#ifdef USE_BM_BACKEND
int crypt_cipher_init(struct crypt_cipher **ctx, const char *name,
		    const char *mode, const void *key, size_t key_length)
{
	*ctx = bm_cipher_open(name, mode, key, key_length);
	return (*ctx)?0:-1;
}

void crypt_cipher_destroy(struct crypt_cipher *ctx)
{
	bm_cipher_close(ctx);
}

int crypt_cipher_encrypt(struct crypt_cipher *ctx,
			 const char *in, char *out, size_t length,
			 const char *iv, size_t iv_length)
{
	//if (iv && ctx->iv_length != iv_length)
	//	return -EINVAL;

	return bm_cipher_encrypt(ctx, (const unsigned char *)in, (unsigned char *)out, length, (const unsigned char *)iv);
}

int crypt_cipher_decrypt(struct crypt_cipher *ctx,
			 const char *in, char *out, size_t length,
			 const char *iv, size_t iv_length)
{
	//if (iv && ctx->iv_length != iv_length)
	//	return -EINVAL;

	return bm_cipher_decrypt(ctx, (const unsigned char *)in, (unsigned char *)out, length, (const unsigned char *)iv);
}

#else
/* Block ciphers */
int crypt_cipher_init(struct crypt_cipher **ctx, const char *name,
		    const char *mode, const void *key, size_t key_length)
{
	struct crypt_cipher *h;
	int r, key_bits;

	h = calloc(1, sizeof(*h));
	if (!h)
		return -ENOMEM;

	key_bits = key_length * 8;
	if (!strcmp(name, "aes"))
	{
		h->iv_length = 16;

		if (!strcmp(mode, "xts"))
			h->chiper_type = CHIPHER_TYPE_AES_XTS;
		else if (!strcmp(mode, "cbc"))
			h->chiper_type = CHIPHER_TYPE_AES_CBC;
		else if (!strcmp(mode, "ecb"))
			h->chiper_type = CHIPHER_TYPE_AES_ECB;
	} else if (!strcmp(name, "serpent"))
	{
		h->iv_length = 16;

		if (!strcmp(mode, "cbc"))
			h->chiper_type = CHIPHER_TYPE_SERPENT_CBC;
		else if (!strcmp(mode, "ecb"))
			h->chiper_type = CHIPHER_TYPE_SERPENT_ECB;
	} else if (!strcmp(name, "twofish"))
	{
		h->iv_length = 16;

		if (!strcmp(mode, "cbc"))
			h->chiper_type = CHIPHER_TYPE_TWOFISH_CBC;
		else if (!strcmp(mode, "ecb"))
			h->chiper_type = CHIPHER_TYPE_TWOFISH_ECB;
	} else if (!strcmp(name, "blowfish"))
	{
		h->iv_length = 8;

		if (!strcmp(mode, "cbc"))
			h->chiper_type = CHIPHER_TYPE_BLOWFISH_CBC;
		else if (!strcmp(mode, "ecb"))
			h->chiper_type = CHIPHER_TYPE_BLOWFISH_ECB;
	} else if (!strcmp(name, "camellia"))
	{
		h->iv_length = 16;

		if (!strcmp(mode, "cbc"))
			h->chiper_type = CHIPHER_TYPE_CAMELLIA_CBC;
		else if (!strcmp(mode, "ecb"))
			h->chiper_type = CHIPHER_TYPE_CAMELLIA_ECB;
	}

	if (h->chiper_type == CHIPHER_TYPE_AES_XTS)
	{
		h->hd_enc = malloc(sizeof(mbedtls_aes_xts_context));
		h->hd_dec = malloc(sizeof(mbedtls_aes_xts_context));
	} else if (h->chiper_type == CHIPHER_TYPE_AES_CBC || h->chiper_type == CHIPHER_TYPE_AES_ECB)
	{
		h->hd_enc = malloc(sizeof(mbedtls_aes_context));
		h->hd_dec = malloc(sizeof(mbedtls_aes_context));
	} else if (h->chiper_type == CHIPHER_TYPE_SERPENT_CBC || h->chiper_type == CHIPHER_TYPE_SERPENT_ECB)
	{
		h->hd_dec = h->hd_enc = malloc(sizeof(mbedtls_serpent_context));
	} else if (h->chiper_type == CHIPHER_TYPE_TWOFISH_CBC || h->chiper_type == CHIPHER_TYPE_TWOFISH_ECB)
	{
		h->hd_dec = h->hd_enc = malloc(sizeof(mbedtls_twofish_context));
	} else if (h->chiper_type == CHIPHER_TYPE_BLOWFISH_CBC || h->chiper_type == CHIPHER_TYPE_BLOWFISH_ECB)
	{
		h->hd_dec = h->hd_enc = malloc(sizeof(mbedtls_blowfish_context));
	} else if (h->chiper_type == CHIPHER_TYPE_CAMELLIA_CBC || h->chiper_type == CHIPHER_TYPE_CAMELLIA_ECB)
	{
		h->hd_enc = malloc(sizeof(mbedtls_camellia_context));
		h->hd_dec = malloc(sizeof(mbedtls_camellia_context));
	} else
	{
		r = -EINVAL;
		goto error;
	}

	if (!h->hd_enc || !h->hd_dec)
	{
		r = -ENOMEM;
		goto error;
	}

	if (h->chiper_type == CHIPHER_TYPE_AES_XTS)
	{
		mbedtls_aes_xts_init((mbedtls_aes_xts_context *)h->hd_enc);
		mbedtls_aes_xts_init((mbedtls_aes_xts_context *)h->hd_dec);
		
		r = mbedtls_aes_xts_setkey_enc((mbedtls_aes_xts_context *)h->hd_enc, key, key_bits) ||
			mbedtls_aes_xts_setkey_dec((mbedtls_aes_xts_context *)h->hd_dec, key, key_bits);
	} else if (h->chiper_type == CHIPHER_TYPE_AES_CBC || h->chiper_type == CHIPHER_TYPE_AES_ECB)
	{
		mbedtls_aes_init((mbedtls_aes_context *)h->hd_enc);
		mbedtls_aes_init((mbedtls_aes_context *)h->hd_dec);

		r = mbedtls_aes_setkey_enc((mbedtls_aes_context *)h->hd_enc, key, key_bits) ||
			mbedtls_aes_setkey_dec((mbedtls_aes_context *)h->hd_dec, key, key_bits);
	} else if (h->chiper_type == CHIPHER_TYPE_SERPENT_CBC || h->chiper_type == CHIPHER_TYPE_SERPENT_ECB)
	{
		mbedtls_serpent_init((mbedtls_serpent_context *)h->hd_enc);
		r = mbedtls_serpent_setup((mbedtls_serpent_context *)h->hd_enc, key, key_bits);
	} else if (h->chiper_type == CHIPHER_TYPE_TWOFISH_CBC || h->chiper_type == CHIPHER_TYPE_TWOFISH_ECB)
	{
		mbedtls_twofish_init((mbedtls_twofish_context *)h->hd_enc);
		r = mbedtls_twofish_setup((mbedtls_twofish_context *)h->hd_enc, key, key_bits);
	} else if (h->chiper_type == CHIPHER_TYPE_BLOWFISH_CBC || h->chiper_type == CHIPHER_TYPE_BLOWFISH_ECB)
	{
		mbedtls_blowfish_init((mbedtls_blowfish_context *)h->hd_enc);
		r = mbedtls_blowfish_setkey((mbedtls_blowfish_context *)h->hd_enc, key, key_bits);
	} else if (h->chiper_type == CHIPHER_TYPE_CAMELLIA_CBC || h->chiper_type == CHIPHER_TYPE_CAMELLIA_ECB)
	{
		mbedtls_camellia_init((mbedtls_camellia_context *)h->hd_enc);
		mbedtls_camellia_init((mbedtls_camellia_context *)h->hd_dec);

		r = mbedtls_camellia_setkey_enc((mbedtls_camellia_context *)h->hd_enc, key, key_bits) ||
			mbedtls_camellia_setkey_dec((mbedtls_camellia_context *)h->hd_dec, key, key_bits);
	} else
		r = 1;

	if (r)
	{
		r = -EINVAL;
		goto error;
	}

	*ctx = h;
	return 0;

error:
	if (h)
	{
		if (h->hd_enc)
			free(h->hd_enc);
		if (h->hd_dec && h->hd_enc != h->hd_dec)
			free(h->hd_dec);
		free(h);		
	}
	
	return r;
}

void crypt_cipher_destroy(struct crypt_cipher *ctx)
{
	if (ctx->chiper_type == CHIPHER_TYPE_AES_XTS)
	{
		mbedtls_aes_xts_free((mbedtls_aes_xts_context *)ctx->hd_enc);
		mbedtls_aes_xts_free((mbedtls_aes_xts_context *)ctx->hd_dec);
	} else if (ctx->chiper_type == CHIPHER_TYPE_AES_CBC || ctx->chiper_type == CHIPHER_TYPE_AES_ECB)
	{
		mbedtls_aes_free((mbedtls_aes_context *)ctx->hd_enc);
		mbedtls_aes_free((mbedtls_aes_context *)ctx->hd_dec);
	} else if (ctx->chiper_type == CHIPHER_TYPE_SERPENT_CBC || ctx->chiper_type == CHIPHER_TYPE_SERPENT_ECB)
	{
		mbedtls_serpent_free((mbedtls_serpent_context *)ctx->hd_enc);
	} else if (ctx->chiper_type == CHIPHER_TYPE_TWOFISH_CBC || ctx->chiper_type == CHIPHER_TYPE_TWOFISH_ECB)
	{
		mbedtls_twofish_free((mbedtls_twofish_context *)ctx->hd_enc);
	} else if (ctx->chiper_type == CHIPHER_TYPE_BLOWFISH_CBC || ctx->chiper_type == CHIPHER_TYPE_BLOWFISH_ECB)
	{
		mbedtls_blowfish_free((mbedtls_blowfish_context *)ctx->hd_enc);
	} else if (ctx->chiper_type == CHIPHER_TYPE_CAMELLIA_CBC || ctx->chiper_type == CHIPHER_TYPE_CAMELLIA_ECB)
	{
		mbedtls_camellia_free((mbedtls_camellia_context *)ctx->hd_enc);
		mbedtls_camellia_free((mbedtls_camellia_context *)ctx->hd_dec);
	}
	
	free(ctx->hd_enc);
	if (ctx->hd_enc != ctx->hd_dec)
		free(ctx->hd_dec);
	free(ctx);
}

int crypt_cipher_encrypt(struct crypt_cipher *ctx,
			 const char *in, char *out, size_t length,
			 const char *iv, size_t iv_length)
{
	int r;

	if (iv && ctx->iv_length != iv_length)
		return -EINVAL;

	switch(ctx->chiper_type)
	{
	case CHIPHER_TYPE_AES_XTS:
		r = mbedtls_aes_crypt_xts((mbedtls_aes_xts_context *)ctx->hd_enc,
                           MBEDTLS_AES_ENCRYPT, length, (const unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_AES_CBC:
		r = mbedtls_aes_crypt_cbc((mbedtls_aes_context *)ctx->hd_enc,
                           MBEDTLS_AES_ENCRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_AES_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_aes_crypt_ecb((mbedtls_aes_context *)ctx->hd_enc,
							   MBEDTLS_AES_ENCRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_SERPENT_CBC:
		r = mbedtls_serpent_crypt_cbc((mbedtls_serpent_context *)ctx->hd_enc,
                           MBEDTLS_SERPENT_ENCRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_SERPENT_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_serpent_crypt_ecb((mbedtls_serpent_context *)ctx->hd_enc,
							   MBEDTLS_SERPENT_ENCRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_TWOFISH_CBC:
		r = mbedtls_twofish_crypt_cbc((mbedtls_twofish_context *)ctx->hd_enc,
                           MBEDTLS_TWOFISH_ENCRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_TWOFISH_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_twofish_crypt_ecb((mbedtls_twofish_context *)ctx->hd_enc,
							   MBEDTLS_TWOFISH_ENCRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_BLOWFISH_CBC:
		r = mbedtls_blowfish_crypt_cbc((mbedtls_blowfish_context *)ctx->hd_enc,
                           MBEDTLS_BLOWFISH_ENCRYPT, length, (unsigned char *)iv,
                           (const unsigned char *)in, (unsigned char *)out);
        break;
	case CHIPHER_TYPE_BLOWFISH_ECB:
		for (int i=0; i<length; i+=8)
		{
			r = mbedtls_blowfish_crypt_ecb((mbedtls_blowfish_context *)ctx->hd_enc,
							   MBEDTLS_BLOWFISH_ENCRYPT, 
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_CAMELLIA_CBC:
		r = mbedtls_camellia_crypt_cbc((mbedtls_camellia_context *)ctx->hd_enc,
                           MBEDTLS_CAMELLIA_ENCRYPT, length, (unsigned char *)iv,
                           (const unsigned char *)in, (unsigned char *)out);
        break;
	case CHIPHER_TYPE_CAMELLIA_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_camellia_crypt_ecb((mbedtls_camellia_context *)ctx->hd_enc,
							   MBEDTLS_CAMELLIA_ENCRYPT, 
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	default: return -EINVAL;
	}

	if (r)
		return -EINVAL;

	return 0;
}

int crypt_cipher_decrypt(struct crypt_cipher *ctx,
			 const char *in, char *out, size_t length,
			 const char *iv, size_t iv_length)
{
	int r;

	if (iv && ctx->iv_length != iv_length)
		return -EINVAL;

	switch(ctx->chiper_type)
	{
	case CHIPHER_TYPE_AES_XTS:
		r = mbedtls_aes_crypt_xts((mbedtls_aes_xts_context *)ctx->hd_dec,
                           MBEDTLS_AES_DECRYPT, length, (const unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_AES_CBC:
		r = mbedtls_aes_crypt_cbc((mbedtls_aes_context *)ctx->hd_dec,
                           MBEDTLS_AES_DECRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_AES_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_aes_crypt_ecb((mbedtls_aes_context *)ctx->hd_dec,
							   MBEDTLS_AES_DECRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_SERPENT_CBC:
		r = mbedtls_serpent_crypt_cbc((mbedtls_serpent_context *)ctx->hd_dec,
                           MBEDTLS_SERPENT_DECRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_SERPENT_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_serpent_crypt_ecb((mbedtls_serpent_context *)ctx->hd_dec,
							   MBEDTLS_SERPENT_DECRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_TWOFISH_CBC:
		r = mbedtls_twofish_crypt_cbc((mbedtls_twofish_context *)ctx->hd_dec,
                           MBEDTLS_TWOFISH_DECRYPT, length, (unsigned char *)iv, 
                           (const unsigned char *)in, (unsigned char *)out);
		break;
	case CHIPHER_TYPE_TWOFISH_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_twofish_crypt_ecb((mbedtls_twofish_context *)ctx->hd_dec,
							   MBEDTLS_TWOFISH_DECRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_BLOWFISH_CBC:
		r = mbedtls_blowfish_crypt_cbc((mbedtls_blowfish_context *)ctx->hd_dec,
                           MBEDTLS_BLOWFISH_DECRYPT, length, (unsigned char *)iv,
                           (const unsigned char *)in, (unsigned char *)out);
        break;
	case CHIPHER_TYPE_BLOWFISH_ECB:
		for (int i=0; i<length; i+=8)
		{
			r = mbedtls_blowfish_crypt_ecb((mbedtls_blowfish_context *)ctx->hd_dec,
							   MBEDTLS_BLOWFISH_DECRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	case CHIPHER_TYPE_CAMELLIA_CBC:
		r = mbedtls_camellia_crypt_cbc((mbedtls_camellia_context *)ctx->hd_dec,
                           MBEDTLS_CAMELLIA_DECRYPT, length, (unsigned char *)iv,
                           (const unsigned char *)in, (unsigned char *)out);
        break;
	case CHIPHER_TYPE_CAMELLIA_ECB:
		for (int i=0; i<length; i+=16)
		{
			r = mbedtls_camellia_crypt_ecb((mbedtls_camellia_context *)ctx->hd_dec,
							   MBEDTLS_CAMELLIA_DECRYPT,
							   (const unsigned char *)in+i, (unsigned char *)out+i);
			if (r)
				break;
		}
		break;
	default: return -EINVAL;
	}

	if (r)
		return -EINVAL;

	return 0;
}
#endif

bool crypt_cipher_kernel_only(struct crypt_cipher *ctx)
{
	return false;
}

int crypt_bitlk_decrypt_key(const void *key, size_t key_length,
			    const char *in, char *out, size_t length,
			    const char *iv, size_t iv_length,
			    const char *tag, size_t tag_length)
{
	mbedtls_ccm_context ctx;
	mbedtls_ccm_init(&ctx);

	if (mbedtls_ccm_setkey(&ctx, MBEDTLS_CIPHER_ID_AES, key, 8*key_length))
		return -EINVAL;

	int r = mbedtls_ccm_auth_decrypt(&ctx, length, 
										(const unsigned char*)iv, iv_length, NULL, 0,
										(const unsigned char*)in, (unsigned char*)out, 
										(const unsigned char*)tag, tag_length);

	mbedtls_ccm_free(&ctx);
	return r;
}

int crypt_backend_memeq(const void *m1, const void *m2, size_t n)
{
	return crypt_internal_memeq(m1, m2, n);
}

bool crypt_fips_mode(void)
{
	return false;
}
