/*
 * Copyright (C) 2021 Ant-ON <prozanton@gmail.com>
 *
 * Based on the Linux kernel DM driver
 *
 * This file is released under the GPL.
 */

#include <endian.h>
#include <stdint.h>
#include <stdlib.h>

#include <bm.h>
#include <bm_module.h>

#include <mbedtls/aes.h>

#include "src/bitops.h"

/*
 * diffuser implementation from Linux kernel 
 * https://github.com/torvalds/linux/blob/v5.6/drivers/md/dm-crypt.c
 */

static void diffuser_disk_to_cpu(uint32_t *d, size_t n)
{
#if __BYTE_ORDER != __LITTLE_ENDIAN
	int i;

	for (i = 0; i < n; i++)
		d[i] = le32_to_cpu((__le32)d[i]);
#endif
}

static void diffuser_cpu_to_disk(uint32_t *d, size_t n)
{
#if __BYTE_ORDER != __LITTLE_ENDIAN
	int i;

	for (i = 0; i < n; i++)
		d[i] = cpu_to_le32((uint32_t)d[i]);
#endif
}

static void diffuser_a_decrypt(uint32_t *d, size_t n)
{
	int i, i1, i2, i3;

	for (i = 0; i < 5; i++) {
		i1 = 0;
		i2 = n - 2;
		i3 = n - 5;

		while (i1 < (n - 1)) {
			d[i1] += d[i2] ^ (d[i3] << 9 | d[i3] >> 23);
			i1++; i2++; i3++;

			if (i3 >= n)
				i3 -= n;

			d[i1] += d[i2] ^ d[i3];
			i1++; i2++; i3++;

			if (i2 >= n)
				i2 -= n;

			d[i1] += d[i2] ^ (d[i3] << 13 | d[i3] >> 19);
			i1++; i2++; i3++;

			d[i1] += d[i2] ^ d[i3];
			i1++; i2++; i3++;
		}
	}
}

static void diffuser_a_encrypt(uint32_t *d, size_t n)
{
	int i, i1, i2, i3;

	for (i = 0; i < 5; i++) {
		i1 = n - 1;
		i2 = n - 2 - 1;
		i3 = n - 5 - 1;

		while (i1 > 0) {
			d[i1] -= d[i2] ^ d[i3];
			i1--; i2--; i3--;

			d[i1] -= d[i2] ^ (d[i3] << 13 | d[i3] >> 19);
			i1--; i2--; i3--;

			if (i2 < 0)
				i2 += n;

			d[i1] -= d[i2] ^ d[i3];
			i1--; i2--; i3--;

			if (i3 < 0)
				i3 += n;

			d[i1] -= d[i2] ^ (d[i3] << 9 | d[i3] >> 23);
			i1--; i2--; i3--;
		}
	}
}

static void diffuser_b_decrypt(uint32_t *d, size_t n)
{
	int i, i1, i2, i3;

	for (i = 0; i < 3; i++) {
		i1 = 0;
		i2 = 2;
		i3 = 5;

		while (i1 < (n - 1)) {
			d[i1] += d[i2] ^ d[i3];
			i1++; i2++; i3++;

			d[i1] += d[i2] ^ (d[i3] << 10 | d[i3] >> 22);
			i1++; i2++; i3++;

			if (i2 >= n)
				i2 -= n;

			d[i1] += d[i2] ^ d[i3];
			i1++; i2++; i3++;

			if (i3 >= n)
				i3 -= n;

			d[i1] += d[i2] ^ (d[i3] << 25 | d[i3] >> 7);
			i1++; i2++; i3++;
		}
	}
}

static void diffuser_b_encrypt(uint32_t *d, size_t n)
{
	int i, i1, i2, i3;

	for (i = 0; i < 3; i++) {
		i1 = n - 1;
		i2 = 2 - 1;
		i3 = 5 - 1;

		while (i1 > 0) {
			d[i1] -= d[i2] ^ (d[i3] << 25 | d[i3] >> 7);
			i1--; i2--; i3--;

			if (i3 < 0)
				i3 += n;

			d[i1] -= d[i2] ^ d[i3];
			i1--; i2--; i3--;

			if (i2 < 0)
				i2 += n;

			d[i1] -= d[i2] ^ (d[i3] << 10 | d[i3] >> 22);
			i1--; i2--; i3--;

			d[i1] -= d[i2] ^ d[i3];
			i1--; i2--; i3--;
		}
	}
}

ssize_t fe_encrypt(void *flt, void *buf, size_t size, off_t sector)
{
	mbedtls_aes_context *enc = ((crypt_filter_t*)flt)->data;
	
	uint8_t es[16] __attribute__((aligned(8)));
	uint8_t ks[32] __attribute__((aligned(4)));

	*((int64_t*)es) = cpu_to_le64(sector*size);
	*((int64_t*)(es+8)) = 0;

	mbedtls_aes_crypt_ecb(enc, MBEDTLS_AES_ENCRYPT, es, ks);
	es[15] = 0x80;
	mbedtls_aes_crypt_ecb(enc, MBEDTLS_AES_ENCRYPT, es, &ks[16]);

	uint32_t *buffer = (uint32_t*)buf;
	size_t size32 = size / sizeof(uint32_t);

	uint32_t *ks32 = (uint32_t*)ks;
	for(int i = 0; i < size32; ++i)
		buffer[i] ^= ks32[i % 8];

	diffuser_disk_to_cpu(buffer, size32);
	diffuser_a_encrypt(buffer, size32);
	diffuser_b_encrypt(buffer, size32);
	diffuser_cpu_to_disk(buffer, size32);

	return size;
}

ssize_t fe_decrypt(void *flt, void *buf, size_t size, off_t sector)
{
	mbedtls_aes_context *enc = ((crypt_filter_t*)flt)->data;

	uint8_t es[16] __attribute__((aligned(8)));
	uint8_t ks[32] __attribute__((aligned(4)));

	*((int64_t*)es) = cpu_to_le64(sector*size);
	*((int64_t*)(es+8)) = 0;

	mbedtls_aes_crypt_ecb(enc, MBEDTLS_AES_ENCRYPT, es, ks);
	es[15] = 0x80;
	mbedtls_aes_crypt_ecb(enc, MBEDTLS_AES_ENCRYPT, es, &ks[16]);

	uint32_t *buffer = (uint32_t*)buf;
	size_t size32 = size / sizeof(uint32_t);

	diffuser_disk_to_cpu(buffer, size32);
	diffuser_b_decrypt(buffer, size32);
	diffuser_a_decrypt(buffer, size32);
	diffuser_cpu_to_disk(buffer, size32);

	uint32_t *ks32 = (uint32_t*)ks;
	for(int i = 0; i < size32; ++i)
		buffer[i] ^= ks32[i % 8];

	return size;
}

void fe_freee(void *flt)
{
	crypt_filter_t *f = flt;
	
	if (f->data)
		free(f->data);
	free(f);
}

crypt_filter_t* filter_elephant_init(const void *key, size_t key_length)
{
	crypt_filter_t *flt = malloc(sizeof(crypt_filter_t));
	flt->encrypt = fe_encrypt;
	flt->decrypt = fe_decrypt;
	flt->freee = fe_freee;
	flt->data = malloc(sizeof(mbedtls_aes_context));

	mbedtls_aes_context *enc = flt->data;
	mbedtls_aes_init(enc);
	mbedtls_aes_setkey_enc(enc, key, key_length*8);
	return flt;
}
