/*
 * libdevmapper - device-mapper backend for cryptsetup
 *
 * Copyright (C) 2004 Jana Saout <jana@saout.de>
 * Copyright (C) 2004-2007 Clemens Fruhwirth <clemens@endorphin.org>
 * Copyright (C) 2009-2020 Red Hat, Inc. All rights reserved.
 * Copyright (C) 2009-2020 Milan Broz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <stdbool.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>

#include "internal.h"

#include <cryptsetup-java.h>
#include <log.h>

/* This doesn't run any kernel checks, just set up userspace libdevmapper */
void dm_backend_init(struct crypt_device *cd)
{
}

void dm_backend_exit(struct crypt_device *cd)
{
}

const char *dm_get_dir(void)
{
	return "";
}

int dm_flags(struct crypt_device *cd, dm_target_type target, uint32_t *flags)
{
	*flags = DM_PLAIN64_SUPPORTED | DM_SECTOR_SIZE_SUPPORTED |
		CRYPT_ACTIVATE_NO_UUID | CRYPT_ACTIVATE_IV_LARGE_SECTORS | CRYPT_ACTIVATE_READONLY | 
		DM_BITLK_EBOIV_SUPPORTED | DM_BITLK_ELEPHANT_SUPPORTED;
	return 0;
}

extern crypt_filter_t* filter_elephant_init(const void *key, size_t key_length);

int dm_create_device(struct crypt_device *cd, const char *name,
		     const char *type,
		     struct crypt_dm_active_device *dmd)
{
	uint32_t dmt_flags;
	int r = -EINVAL;

	if (!type || !dmd)
		return -EINVAL;

	if (dmd->flags & CRYPT_ACTIVATE_PRIVATE)
	{
		log_err(cd, "libbm not supported private device");
		return -EINVAL; // Don't support private device
	}

	struct and_device *dev = (struct and_device *)crypt_get_userptr(cd);
	if (!dev)
		return -EINVAL;	

	dm_flags(cd, 0, &dmt_flags);
	if (dmd->flags&(~dmt_flags))
	{
		log_err(cd, "libbm not supported flags 0x%X", (dmd->flags&(~dmt_flags)));
		dev->flags |= FLAG_UNSUPPORTED_FEATURE;
		return -EINVAL;
	}

	struct dm_target *first = &dmd->segment;
	if (first->type != DM_CRYPT)
	{
		// BM support multi-segment device with start from crypto
		log_err(cd, "libbm not supported start from type other than DM_CRYPT");
		dev->flags |= FLAG_UNSUPPORTED_FEATURE;
		return -EINVAL;
	}

	struct dm_target *s = &dmd->segment;
	while (s)
	{
		if (s->type == DM_CRYPT)
		{
			log_std(cd, "%15" PRIu64 " %15" PRIu64 " %15" PRIu64 " C %11" PRIu64, s->offset, (s->offset+s->size), s->size, s->u.crypt.offset);

			if (strcmp(first->u.crypt.cipher, s->u.crypt.cipher))
			{
				log_err(cd, "libbm not supported DM_CRYPT segments with different chipers");
				dev->flags |= FLAG_UNSUPPORTED_FEATURE;
				return -EINVAL;
			} else if (first->u.crypt.vk->keylength != s->u.crypt.vk->keylength ||
					memcmp(first->u.crypt.vk->key, s->u.crypt.vk->key, s->u.crypt.vk->keylength))
			{
				log_err(cd, "libbm not supported DM_CRYPT segments with different key");
				dev->flags |= FLAG_UNSUPPORTED_FEATURE;
				return -EINVAL;
			} else if (first->u.crypt.sector_size != s->u.crypt.sector_size)
			{
				log_err(cd, "libbm not supported DM_CRYPT segments with different sector size");
				dev->flags |= FLAG_UNSUPPORTED_FEATURE;
				return -EINVAL;
			}
		} else if (s->type == DM_ZERO)
		{
			log_std(cd, "%15" PRIu64 " %15" PRIu64 " %15" PRIu64 " Z", s->offset, (s->offset+s->size), s->size);
		} else
		{
			log_err(cd, "libbm not supported segments with %d type", s->type);
			dev->flags |= FLAG_UNSUPPORTED_FEATURE;
			return -EINVAL;
		}
		
		s = s->next;
	}

	s = &dmd->segment;

	crypt_filter_t *flt = NULL;
	size_t keylength = s->u.crypt.vk->keylength;
	const char *chipher = s->u.crypt.cipher;
	if (!strcmp(chipher, "aes-cbc-elephant"))
	{
		// Configure data filter (converter) for implement elephant IV
		chipher = "aes-cbc-eboiv";
		keylength = keylength/2;
		flt = filter_elephant_init(s->u.crypt.vk->key+keylength, keylength);
	}

	r = bm_crypt_init(&dev->bm_blk, dev->blk,
			s->offset*SECTOR_SIZE, s->size*SECTOR_SIZE,
			s->u.crypt.offset,
			s->u.crypt.iv_offset,
			s->u.crypt.sector_size,
			chipher,
			s->u.crypt.vk->key, keylength, flt, dmd->flags);
	if (r || !dev->bm_blk)
		goto error;

	s = s->next;
	while (s)
	{
		if (s->type == DM_CRYPT)
		{
			r = bm_crypt_add_crypt(dev->bm_blk,
								s->offset, s->size,
								s->u.crypt.offset,
								s->u.crypt.iv_offset);
		} else if (s->type == DM_ZERO)
		{
			r = bm_crypt_add_zero(dev->bm_blk,
								s->offset, s->size);
		}
		
		if (r)
			goto error;
		s = s->next;
	}

	return 0;
	
error:
	if (dev->bm_blk)
	{
		dev->bm_blk->freee(dev->bm_blk);
		dev->bm_blk = NULL;
	}
	return r;
}

int dm_remove_device(struct crypt_device *cd, const char *name, uint32_t flags)
{
	return -ENOTSUP;
}

int dm_query_device(struct crypt_device *cd, const char *name,
		    uint32_t get_flags, struct crypt_dm_active_device *dmd)
{
	return -ENOTSUP;
}

int dm_resume_device(struct crypt_device *cd, const char *name, uint32_t dmflags)
{
	return -ENOTSUP;
}

int dm_reload_device(struct crypt_device *cd, const char *name,
		     struct crypt_dm_active_device *dmd, uint32_t dmflags, unsigned resume)
{
	return -ENOTSUP;
}

int dm_status_device(struct crypt_device *cd, const char *name)
{
	return -ENODEV;
}

int dm_status_suspended(struct crypt_device *cd, const char *name)
{
	return -ENOTSUP;
}

int dm_device_deps(struct crypt_device *cd, const char *name, const char *prefix, char **names, size_t names_length)
{
	return -ENODEV;
}

int dm_crypt_target_set(struct dm_target *tgt, uint64_t seg_offset, uint64_t seg_size,
	struct device *data_device, struct volume_key *vk, const char *cipher,
	uint64_t iv_offset, uint64_t data_offset, const char *integrity, uint32_t tag_size,
	uint32_t sector_size)
{
	int r = -EINVAL;

	/* free on error */
	char *dm_integrity = NULL;

	if (tag_size) {
		/* Space for IV metadata only */
		dm_integrity = strdup(integrity ?: "none");
		if (!dm_integrity) {
			r = -ENOMEM;
			goto err;
		}
	}

	tgt->data_device = data_device;

	tgt->type = DM_CRYPT;
	tgt->direction = TARGET_SET;
	tgt->u.crypt.vk = vk;
	tgt->offset = seg_offset;
	tgt->size = seg_size;

	tgt->u.crypt.cipher = cipher;
	tgt->u.crypt.integrity = dm_integrity;
	tgt->u.crypt.iv_offset = iv_offset;
	tgt->u.crypt.offset = data_offset;
	tgt->u.crypt.tag_size = tag_size;
	tgt->u.crypt.sector_size = sector_size;

	return 0;
err:
	free(dm_integrity);

	return r;
}

int dm_zero_target_set(struct dm_target *tgt, uint64_t seg_offset, uint64_t seg_size)
{
	tgt->type = DM_ZERO;
	tgt->direction = TARGET_SET;
	tgt->offset = seg_offset;
	tgt->size = seg_size;

	return 0;
}

int dm_linear_target_set(struct dm_target *tgt, uint64_t seg_offset, uint64_t seg_size,
	struct device *data_device, uint64_t data_offset)
{
	return -ENOTSUP;
}

int dm_integrity_target_set(struct crypt_device *cd,
			struct dm_target *tgt, uint64_t seg_offset, uint64_t seg_size,
			struct device *meta_device,
		        struct device *data_device, uint64_t tag_size, uint64_t offset,
			uint32_t sector_size, struct volume_key *vk,
			struct volume_key *journal_crypt_key, struct volume_key *journal_mac_key,
			const struct crypt_params_integrity *ip)
{
	return -ENOTSUP;
}

int dm_verity_target_set(struct dm_target *tgt, uint64_t seg_offset, uint64_t seg_size,
	struct device *data_device, struct device *hash_device, struct device *fec_device,
	const char *root_hash, uint32_t root_hash_size, const char *root_hash_sig_key_desc,
	uint64_t hash_offset_block, uint64_t hash_blocks, struct crypt_params_verity *vp)
{
	return -ENOTSUP;
}

static void _dm_target_free_query_path(struct crypt_device *cd, struct dm_target *tgt)
{
	switch(tgt->type) {
	case DM_CRYPT:
		crypt_free_volume_key(tgt->u.crypt.vk);
		free(CONST_CAST(void*)tgt->u.crypt.cipher);
		break;
	case DM_INTEGRITY:
		/* dm_integrity_target_set not use */
		break;
	case DM_VERITY:
		/* dm_verity_target_set not use */
	case DM_LINEAR:
		/* fall through */
	case DM_ERROR:
		/* fall through */
	case DM_ZERO:
		break;
	default:
		log_err(cd, "Unknown dm target type.");
		return;
	}

	device_free(cd, tgt->data_device);
}

static void _dm_target_erase(struct crypt_device *cd, struct dm_target *tgt)
{
	if (tgt->direction == TARGET_QUERY)
		_dm_target_free_query_path(cd, tgt);

	if (tgt->type == DM_CRYPT)
		free(CONST_CAST(void*)tgt->u.crypt.integrity);
}

void dm_targets_free(struct crypt_device *cd, struct crypt_dm_active_device *dmd)
{
	struct dm_target *t = &dmd->segment, *next = t->next;

	_dm_target_erase(cd, t);

	while (next) {
		t = next;
		next = t->next;
		_dm_target_erase(cd, t);
		free(t);
	}

	memset(&dmd->segment, 0, sizeof(dmd->segment));
}

int dm_targets_allocate(struct dm_target *first, unsigned count)
{
	if (!first || first->next || !count)
		return -EINVAL;

	while (--count) {
		first->next = crypt_zalloc(sizeof(*first));
		if (!first->next)
			return -ENOMEM;
		first = first->next;
	}

	return 0;
}

int lookup_dm_dev_by_uuid(struct crypt_device *cd, const char *uuid, const char *type)
{
	return 0; /* not found */
}

int dm_suspend_device(struct crypt_device *cd, const char *name, uint32_t dmflags)
{
	return -ENOTSUP;
}

int dm_error_device(struct crypt_device *cd, const char *name)
{
	return -ENOTSUP;
}

int dm_clear_device(struct crypt_device *cd, const char *name)
{
	return -ENOTSUP;
}

int dm_status_verity_ok(struct crypt_device *cd, const char *name)
{
	return -ENOTSUP;
}	

int dm_resume_and_reinstate_key(struct crypt_device *cd, const char *name,
				const struct volume_key *vk)
{
	return -ENOTSUP;
}

int dm_cancel_deferred_removal(const char *name)
{
	return -ENOTSUP;
}

int dm_status_integrity_failures(struct crypt_device *cd, const char *name, uint64_t *count)
{
	return -ENOTSUP;
}

char *dm_device_name(const char *path)
{
	return NULL;
}
